__all__ = ()

import warnings
from itertools import chain

try:
    from _weakref import WeakSet
except ImportError:
    from weakref import WeakSet

from ...backend.export import export
from ...backend.futures import Task

from ..core import KOKORO

from .intent import DISPATCH_EVENT_TO_INTENTS, INTENT_SHIFT_MISSING_EVENT, INTENT_SHIFT_DEFAULT_EVENT

PARSERS = {}

EVENT_HANDLER_EXPECTED_ARGUMENT_COUNTS = {}
EVENT_HANDLER_NAME_TO_PARSER_NAMES = {}
EVENT_HANDLER_NAMES = set()

REGISTERED_CLIENTS = WeakSet()
PARSER_SETTINGS = {}

async def DEFAULT_EVENT_HANDLER(*args):
    """
    Default event handler what is set under events if there is no specified event handler to use.
    
    This function is a coroutine.
    
    Parameters
    ----------
    *args : Positional parameters
    """
    pass

def _iter_name(name):
    """
    Iterates the given dispatch event name.
    
    This function is a generator.
    
    Parameters
    ----------
    name : `str` or `tuple` of (`str` or ...)
        The name or names of the dispatch events.
    
    Yields
    ------
    name : `str`
    """
    if name is None:
        pass
    elif isinstance(name, tuple):
        for name in name:
            yield from _iter_name(name)
    else:
        yield name


def add_event_handler(name, value, parser):
    """
    Adds a new event-name argcount parser names relation to the event system core instance.
    
    Parameters
    ----------
    name : `str`
        The name of the event.
    value : `int`
        The amount of parameters what the parser passes the respective event.
    parser : `str` or (`tuple` of `str`)
        The name of events, which might call the respective event.
    """
    parser = tuple(_iter_name(parser))
    
    EVENT_HANDLER_EXPECTED_ARGUMENT_COUNTS[name] = value
    EVENT_HANDLER_NAME_TO_PARSER_NAMES[name] = parser
    EVENT_HANDLER_NAMES.add(name)


def get_plugin_event_handler_and_parameter_count(event_handler_manager, name):
    """
    Returns the amount of parameters, what the events would pass to the respective event.
    
    Parameters
    ----------
    event_handler_manager : ``EventHandlerManager``
        The respective event handler.
    name : `str`
        The event's name.
    
    Returns
    -------
    plugin : ``EventHandlerManager``, ``EventHandlerPlugin``
        The event handler or the plugin owning the event.
    parameter_count : `int`
        The amount of parameters, what to the respective event would be passed.
    """
    try:
        parameter_count = EVENT_HANDLER_EXPECTED_ARGUMENT_COUNTS[name]
    except KeyError:
        pass
    else:
        return event_handler_manager, parameter_count
    
    plugin_events = event_handler_manager._plugin_events
    if (plugin_events is not None):
        try:
            event_handler_plugin = plugin_events[name]
        except KeyError:
            pass
        else:
            return event_handler_plugin, event_handler_plugin._plugin_parameter_counts[name]
    
    return None, 0


def get_plugin_event_handler_and_parser_names(event_handler_manager, name):
    """
    Returns the plugin implementing the given attribute, and it's dispatch event parser names.
    
    Parameters
    ----------
    event_handler_manager : ``EventHandlerManager``
        The respective event handler.
    name : `str`
        The event's name.
    
    Returns
    -------
    plugin : `None`, ``EventHandlerManager``, ``EventHandlerPlugin``
        The event handler or the plugin owning the event.
    parser_names : `None` or `tuple` of `str`
        Dispatch event parser's names relating to the event.
    """
    try:
        parser_names = EVENT_HANDLER_NAME_TO_PARSER_NAMES[name]
    except KeyError:
        pass
    else:
        return event_handler_manager, parser_names
    
    plugin_events = event_handler_manager._plugin_events
    if (plugin_events is not None):
        try:
            event_handler_plugin = plugin_events[name]
        except KeyError:
            pass
        else:
            return event_handler_plugin, None
    
    return None, None


def get_plugin_event_handler(event_handler_manager, name):
    """
    Returns the plugin implementing the attribute.
    
    Parameters
    ----------
    event_handler_manager : ``EventHandlerManager``
        The respective event handler.
    name : `str`
        The event's name.
    
    Returns
    -------
    plugin : `None`, ``EventHandlerManager``, ``EventHandlerPlugin``
        The event handler or the plugin owning the event.
    """
    if name in EVENT_HANDLER_NAMES:
        return event_handler_manager
    
    plugin_events = event_handler_manager._plugin_events
    if (plugin_events is not None):
        try:
            event_handler_plugin = plugin_events[name]
        except KeyError:
            pass
        else:
            return event_handler_plugin
    
    return None


add_event_handler('error', 3, None,)
add_event_handler('launch', 1, None,)
add_event_handler('shutdown', 1, None,)
add_event_handler('voice_client_ghost', 2, None,)
add_event_handler('voice_client_join', 2, None,)
add_event_handler('voice_client_move', 3, None,)
add_event_handler('voice_client_leave', 3, None,)
add_event_handler('voice_client_update', 3, None,)
add_event_handler('voice_client_shutdown', 1, None,)

add_event_handler('ready', 1, 'READY',)
add_event_handler('client_edit', 2, 'USER_UPDATE',)
add_event_handler('message_create', 2, 'MESSAGE_CREATE',)
add_event_handler('message_delete', 2, ('MESSAGE_DELETE', 'MESSAGE_DELETE_BULK') ,)
add_event_handler('message_edit', 3, 'MESSAGE_UPDATE',)
add_event_handler('embed_update', 3, 'MESSAGE_UPDATE',)
add_event_handler('reaction_add', 2, 'MESSAGE_REACTION_ADD',)
add_event_handler('reaction_clear', 3, 'MESSAGE_REACTION_REMOVE_ALL',)
add_event_handler('reaction_delete', 2, 'MESSAGE_REACTION_REMOVE',)
add_event_handler('reaction_delete_emoji', 4, 'MESSAGE_REACTION_REMOVE_EMOJI',)
add_event_handler('user_edit', 3, 'PRESENCE_UPDATE',)
add_event_handler('user_presence_update', 3, 'PRESENCE_UPDATE',)
add_event_handler('guild_user_edit', 4, 'GUILD_MEMBER_UPDATE',)
add_event_handler('channel_delete', 2, ('CHANNEL_DELETE', 'THREAD_DELETE'),)
add_event_handler('channel_edit', 3, ('CHANNEL_UPDATE', 'THREAD_UPDATE'),)
add_event_handler('channel_create', 2, ('CHANNEL_CREATE', 'THREAD_CREATE'),)
add_event_handler('channel_pin_update', 2, 'CHANNEL_PINS_UPDATE',)
add_event_handler('channel_group_user_add', 3, 'CHANNEL_RECIPIENT_ADD',)
add_event_handler('channel_group_user_delete', 3, 'CHANNEL_RECIPIENT_REMOVE',)
add_event_handler('emoji_create', 2, 'GUILD_EMOJIS_UPDATE',)
add_event_handler('emoji_delete', 2, 'GUILD_EMOJIS_UPDATE',)
add_event_handler('emoji_edit', 3, 'GUILD_EMOJIS_UPDATE',)
add_event_handler('sticker_create', 2, 'GUILD_STICKERS_UPDATE',)
add_event_handler('sticker_delete', 2, 'GUILD_STICKERS_UPDATE',)
add_event_handler('sticker_edit', 3, 'GUILD_STICKERS_UPDATE',)
add_event_handler('guild_user_add', 3, 'GUILD_MEMBER_ADD',)
add_event_handler('guild_user_delete', 4, 'GUILD_MEMBER_REMOVE',)
add_event_handler('guild_join_reject', 3, 'GUILD_JOIN_REQUEST_DELETE',)
add_event_handler('guild_create', 2, 'GUILD_CREATE',)
add_event_handler('guild_edit', 2, 'GUILD_UPDATE',)
add_event_handler('guild_delete', 3, 'GUILD_DELETE',)
add_event_handler('guild_ban_add', 3, 'GUILD_BAN_ADD',)
add_event_handler('guild_ban_delete', 3, 'GUILD_BAN_REMOVE',)
add_event_handler('guild_user_chunk', 2, 'GUILD_MEMBERS_CHUNK',)
add_event_handler('integration_create', 3, 'INTEGRATION_CREATE',)
add_event_handler('integration_delete', 4, 'INTEGRATION_DELETE',)
add_event_handler('integration_edit', 3, 'INTEGRATION_UPDATE',)
add_event_handler('integration_update', 2, 'GUILD_INTEGRATIONS_UPDATE',)
add_event_handler('role_create', 2, 'GUILD_ROLE_CREATE',)
add_event_handler('role_delete', 3, 'GUILD_ROLE_DELETE',)
add_event_handler('role_edit', 3, 'GUILD_ROLE_UPDATE',)
add_event_handler('webhook_update', 2, 'WEBHOOKS_UPDATE',)
add_event_handler('user_voice_join', 2, 'VOICE_STATE_UPDATE',)
add_event_handler('user_voice_leave', 3, 'VOICE_STATE_UPDATE',)
add_event_handler('user_voice_update', 3, 'VOICE_STATE_UPDATE',)
add_event_handler('user_voice_move', 3, 'VOICE_STATE_UPDATE',)
add_event_handler('typing', 4, 'TYPING_START',)
add_event_handler('invite_create', 2, 'INVITE_CREATE',)
add_event_handler('invite_delete', 2, 'INVITE_DELETE',)
add_event_handler('relationship_add', 2, 'RELATIONSHIP_ADD',)
add_event_handler('relationship_change', 3, 'RELATIONSHIP_ADD',)
add_event_handler('relationship_delete', 2, 'RELATIONSHIP_REMOVE',)
add_event_handler('gift_update', 3, 'GIFT_CODE_UPDATE',)
add_event_handler('interaction_create', 2, 'INTERACTION_CREATE',)
add_event_handler('application_command_create', 3, 'APPLICATION_COMMAND_CREATE',)
add_event_handler('application_command_update', 4, 'APPLICATION_COMMAND_UPDATE',)
add_event_handler('application_command_delete', 3, 'APPLICATION_COMMAND_DELETE',)
add_event_handler('application_command_permission_update', 2, 'APPLICATION_COMMAND_PERMISSIONS_UPDATE',)
add_event_handler('stage_create', 2, 'STAGE_INSTANCE_CREATE',)
add_event_handler('stage_edit', 3, 'STAGE_INSTANCE_UPDATE',)
add_event_handler('stage_delete', 2, 'STAGE_INSTANCE_DELETE',)
add_event_handler('thread_user_add', 4, ('THREAD_MEMBER_UPDATE', 'THREAD_MEMBERS_UPDATE'),)
add_event_handler('thread_user_delete', 3, ('THREAD_MEMBER_UPDATE', 'THREAD_MEMBERS_UPDATE'),)
add_event_handler('voice_server_update', 2, 'VOICE_SERVER_UPDATE',)
add_event_handler('scheduled_event_create', 2, 'GUILD_SCHEDULED_EVENT_CREATE',)
add_event_handler('scheduled_event_edit', 3, 'GUILD_SCHEDULED_EVENT_UPDATE',)
add_event_handler('scheduled_event_delete', 2, 'GUILD_SCHEDULED_EVENT_DELETE',)
add_event_handler('scheduled_event_user_subscribe', 3, 'GUILD_SCHEDULED_EVENT_USER_CREATE',)
add_event_handler('scheduled_event_user_unsubscribe', 3, 'GUILD_SCHEDULED_EVENT_USER_DELETE',)


class ParserSettingOption:
    """
    Contains details about a dispatch event parser
    
    Parameters
    ----------
    intent_shift : `int`
        The event's intent's respective shift.
    name : `str`
        The parser's name(s) also known as the dispatch event's.
    """
    __slots__ = ('intent_shift', 'name')
    
    def __new__(cls, name):
        """
        Creates new ``ParserSettingOption`` instances from the given name.
        
        This method is a generator.
        
        Yields
        ------
        self : ``ParserSettingOption``
        """
        try:
            intent_shifts = DISPATCH_EVENT_TO_INTENTS[name]
        except KeyError:
            warnings.warn(
                f'Dispatch event parser {name!r} is not registered to any intent. '
                'Will always use optimized parser to dispatch it.',
                RuntimeWarning)
            
            intent_shifts = (INTENT_SHIFT_MISSING_EVENT,)
        
        for intent_shift in intent_shifts:
            self = object.__new__(cls)
            self.name = name
            self.intent_shift = intent_shift
            yield self
    
    def __repr__(self):
        """Returns the parser description's representation."""
        return f'<{self.__class__.__name__} name={self.name!r}, intent_shift={self.intent_shift!r}>'


class ParserSetting:
    """
    Stores the events for each dispatch events.
    
    Each dispatch event calls it corresponding parser, what can be 1 of up to 4 different events depending what is the
    optimal way of parsing that specific event. The called parser depends on the running client's intent values and
    whether they have a handler for the respective event. The parser are changed on change, so do not worry, there are
    no useless checks done every time a dispatch event is received.
    
    Attributes
    ----------
    options : `tuple` of ``ParserSettingOption``
        Options about the dispatch events handled by the parser.
    parser_cal_sc : `function`
        Single client parser what calculates the differences between the previous and the current state and calls
        the client's event.
    parser_cal_mc : `function`
        Multi client parser what calculates the differences between the previous and the current state and calls
        the clients' events.
    parser_opt_sc : `function`
        Single client optimized parser.
    parser_opt_mc : `function`
        Multi client optimized events.
    mention_count : `int`
        How much events of the running clients expect to be called by the respective parser. Used for `opt` - `cal`
        optimizations.
    client_count : `int`
        How much running clients expect the respective parser to call their events. Used in `sc` - `mc` optimizations.
    """
    __slots__ = ('options', 'parser_cal_sc',  'parser_opt_sc', 'parser_cal_mc', 'parser_opt_mc', 'mention_count', 'client_count',)
    def __new__(cls, names, parser_cal_sc, parser_cal_mc, parser_opt_sc, parser_opt_mc):
        """
        Creates a new parser defaults object with the given name and with the given events.
        
        The created parser defaults are stored at the class's `.all` attribute and also the default parser, so
        `parser_opt_sc` is set to the global `PARSERS` variable.
        
        Parameters
        ----------
        name : `tuple` of `str`
            The parser's name also known as the dispatch event's.
        parser_cal_sc : `function`
            Single client parser what calculates the differences between the previous and the current state and calls
            the client's event.
        parser_cal_mc : `function`
            Multi client parser what calculates the differences between the previous and the current state and calls
            the clients' events.
        parser_opt_sc : `function`
            Single client optimized parser.
        parser_opt_mc : `function`
            Multi client optimized events.
        """
        options = tuple(chain.from_iterable(ParserSettingOption(name) for name in names))
        
        self = object.__new__(cls)
        self.options = options
        self.parser_cal_sc = parser_cal_sc
        self.parser_cal_mc = parser_cal_mc
        self.parser_opt_sc = parser_opt_sc
        self.parser_opt_mc = parser_opt_mc
        self.mention_count = 0
        self.client_count = 0
        
        return self
    
    def add_mention(self, client):
        """
        If the client is already registered, mentions the respective parser defaults and optimizes the used events.
        
        Parameters
        ----------
        client : ``Client``
        """
        if client is None:
            return
        
        if client not in REGISTERED_CLIENTS:
            return
        
        for option in self.options:
            intent_shift = option.intent_shift
            if (intent_shift == INTENT_SHIFT_DEFAULT_EVENT):
                break
            
            if (client.intents>>intent_shift)&1:
                break
        
        else:
            return
        
        self.mention_count += 1
        self._recalculate()
    
    def remove_mention(self, client):
        """
        If the client is registered to the parser defaults, removes it's mention from the respective parser defaults
        and optimizes the used events.
        
        Parameters
        ----------
        client : ``Client``
        """
        if client is None:
            return
        
        if client not in REGISTERED_CLIENTS:
            return
        
        for description in self.options:
            intent_shift = description.intent_shift
            if (intent_shift == INTENT_SHIFT_DEFAULT_EVENT):
                break
            
            if (client.intents>>intent_shift)&1:
                break
        
        else:
            return
        
        self.mention_count -= 1
        self._recalculate()
    
    def _recalculate(self):
        """
        Chooses the optimal events for each dispatch event.
        """
        mention_count = self.mention_count
        client_count = self.client_count
        
        if mention_count == 0:
            if client_count < 2:
                parser = self.parser_opt_sc
            else:
                parser = self.parser_opt_mc
        else:
            if client_count < 2:
                parser = self.parser_cal_sc
            else:
                parser = self.parser_cal_mc
        
        for option in self.options:
            PARSERS[option.name] = parser


def register_client(client):
    """
    Registers the given client as a running one. It means it's used events will be registered and their
    change will be handled to optimize the used events.
    
    Parameters
    ----------
    client : ``Client``
    """
    REGISTERED_CLIENTS.add(client)
    
    enabled_parsers = set()
    
    if client.is_bot:
        for parser_name in client.intents.iterate_parser_names():
            enabled_parsers.add(parser_name)
    else:
        for parser_name in PARSER_SETTINGS.keys():
            enabled_parsers.add(parser_name)
    
    for parser_name in enabled_parsers:
        parser_default = PARSER_SETTINGS[parser_name]
        parser_default.client_count +=1
        parser_default._recalculate()
    
    for event_name in EVENT_HANDLER_NAME_TO_PARSER_NAMES.keys():
        event = getattr(client.events, event_name)
        if event is DEFAULT_EVENT_HANDLER:
            continue
        
        parser_names = EVENT_HANDLER_NAME_TO_PARSER_NAMES[event_name]
        for parser_name in parser_names:
            if parser_name not in enabled_parsers:
                continue
            
            parser_default = PARSER_SETTINGS[parser_name]
            parser_default.mention_count +=1
            parser_default._recalculate()
            

def unregister_client(client):
    """
    Unregisters the given client, so it's event be unregistered and their change will not be handled anymore to
    optimize the used events.
    
    Parameters
    ----------
    client : ``Client``
    """
    try:
        REGISTERED_CLIENTS.remove(client)
    except ValueError:
        return
    
    enabled_parsers = set()
    
    if client.is_bot:
        for parser_name in client.intents.iterate_parser_names():
            enabled_parsers.add(parser_name)
    else:
        for parser_name in PARSER_SETTINGS.keys():
            enabled_parsers.add(parser_name)
    
    for parser_name in enabled_parsers:
        parser_default = PARSER_SETTINGS[parser_name]
        parser_default.client_count -= 1
        parser_default._recalculate()
    
    for event_name in EVENT_HANDLER_NAME_TO_PARSER_NAMES.keys():
        event = getattr(client.events, event_name)
        if event is DEFAULT_EVENT_HANDLER:
            continue
        
        parser_names = EVENT_HANDLER_NAME_TO_PARSER_NAMES[event_name]
        for parser_name in parser_names:
            if parser_name not in enabled_parsers:
                continue
            
            parser_default = PARSER_SETTINGS[parser_name]
            parser_default.mention_count -= 1
            parser_default._recalculate()
            continue


def add_parser(name, parser_cal_sc, parser_cal_mc, parser_opt_sc, parser_opt_mc):
    """
    Creates a new parser defaults object with the given name and with the given events.
    
    The created parser defaults are stored at the class's `.all` attribute and also the default parser, so
    `parser_opt_sc` is set to the global `PARSERS` variable.
    
    Parameters
    ----------
    name : `str` or `tuple` of (`str`, ...)
        The parser's name also known as the dispatch event's.
    parser_cal_sc : `function`
        Single client parser what calculates the differences between the previous and the current state and calls
        the client's event.
    parser_cal_mc : `function`
        Multi client parser what calculates the differences between the previous and the current state and calls
        the clients' events.
    parser_opt_sc : `function`
        Single client optimized parser.
    parser_opt_mc : `function`
        Multi client optimized events.
    
    Returns
    -------
    parser_setting : ``ParserSetting``
        The registered parser setting.
    """
    names = tuple(_iter_name(name))
    parser_setting = ParserSetting(names, parser_cal_sc, parser_cal_mc, parser_opt_sc, parser_opt_mc)
    
    for option in parser_setting.options:
        name = option.name
        PARSER_SETTINGS[name] = parser_setting
        PARSERS[name] = parser_opt_sc
    
    return parser_setting

def maybe_ensure_launch(client):
    """
    Calls `client.events.launch` if not yet called.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    """
    events = client.events
    if not events._launch_called:
        events._launch_called = True
        
        event_handler = client.events.launch
        if (event_handler is not DEFAULT_EVENT_HANDLER):
            Task(event_handler(client), KOKORO)


@export
def trigger_voice_client_ghost_event(client, voice_state):
    """
    Triggers `Client.events.voice_client_ghost` if set.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    voice_state : ``VoiceState``
        The client's ghost voice state.
    """
    event_handler = client.events.voice_client_ghost
    if (event_handler is not DEFAULT_EVENT_HANDLER):
        Task(event_handler(client, voice_state), KOKORO)



import sys
from platform import platform as get_platform
from .. import __package__ as PACKAGE_NAME, __version__ as VERSION

NAME = 'version'
USAGE = 'v | version'

HELP = (
    f'Prints out hata\'s version.\n'
)

def __main__():
    output_parts = []
    
    output_parts.append(PACKAGE_NAME[0].upper())
    output_parts.append(PACKAGE_NAME[1:])
    output_parts.append(' version: ')
    output_parts.append(VERSION)
    output_parts.append('\n')
    
    output_parts.append('Platform: ')
    output_parts.append(get_platform())
    output_parts.append('\n')
    
    implementation = sys.implementation
    
    output_parts.append('Python implementation: ')
    output_parts.append(implementation.name)
    output_parts.append('\n')
    
    version = implementation.version
    
    output_parts.append('Python version: ')
    output_parts.append(str(version[0]))
    output_parts.append('.')
    output_parts.append(str(version[1]))
    
    if version[3] == 'final':
        output_parts.append(' final')
    
    output_parts.append('\n')
    
    output = ''.join(output_parts)
    
    sys.stdout.write(output)



__all__ = ()

from ...backend.utils import istr

AUDIT_LOG_REASON = istr('X-Audit-Log-Reason')

RATE_LIMIT_REMAINING = istr('X-RateLimit-Remaining')
RATE_LIMIT_RESET = istr('X-RateLimit-Reset')
RATE_LIMIT_RESET_AFTER = istr('X-RateLimit-Reset-After')
RATE_LIMIT_LIMIT = istr('X-RateLimit-Limit')

# to send
RATE_LIMIT_PRECISION = istr('X-RateLimit-Precision')
DEBUG_OPTIONS = istr('X-Debug-Options')



__all__ = ('WebhookBase',)

from ...backend.utils import copy_docs

from ..user import UserBase
from ..core import CHANNELS
from ..permission.permission import PERMISSION_MASK_USE_EXTERNAL_EMOJIS

from .preinstanced import WebhookType

class WebhookBase(UserBase):
    """
    Base class for webhook like-types.
    
    Attributes
    ----------
    id : `int`
        The webhook's unique identifier number.
    name : str
        The webhook's username.
    discriminator : `int`
        The webhook's discriminator. Given to avoid overlapping names.
    avatar_hash : `int`
        The webhook's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The webhook's avatar's type.
    banner_color : `None` or ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    channel_id : `int`
        The channel's identifier, where the webhook is going to send it's messages.
    type : ``WebhookType``
        The webhook's type.
    """
    __slots__ = ('channel_id', 'type',)
    
    @property
    def is_bot(self):
        """
        Webhooks are always bots.
        
        Returns
        -------
        is_bot : `bool`
        """
        return True
    
    
    @property
    def partial(self):
        """
        Returns whether the webhook is partial.
        
        A webhook is partial, if it's respective guild is unknown.
        
        Returns
        -------
        partial : `bool`
        """
        try:
            channel = CHANNELS[self.channel_id]
        except KeyError:
            return True
        
        if channel.guild is None:
            return True
        
        return False
    
    
    @property
    def channel(self):
        """
        Returns the webhook's channel if applicable.
        
        Returns
        -------
        channel : `None` or ``ChannelText``
        """
        return CHANNELS.get(self.channel_id, None)
    
    
    @property
    def guild(self):
        """
        Returns the webhook's guild if applicable.
        
        Returns
        -------
        guild : `None` or ``Guild``
        """
        try:
            channel = CHANNELS[self.channel_id]
        except KeyError:
            pass
        else:
            return channel.guild
    
    
    @copy_docs(UserBase.can_use_emoji)
    def can_use_emoji(self, emoji):
        if emoji.is_unicode_emoji():
            return True
        
        emoji_roles = emoji.emoji_roles
        if (emoji_roles is not None):
            return False
        
        guild = self.guild
        if guild is None:
            return False
        
        default_role = guild.default_role
        if default_role&PERMISSION_MASK_USE_EXTERNAL_EMOJIS:
            return True
        
        return False
    
    
    @copy_docs(UserBase._set_default_attributes)
    def _set_default_attributes(self):
        UserBase._set_default_attributes(self)
        
        self.channel = None
        self.type = WebhookType.none



__all__ = ('CommandCooldownWrapper', 'CommandCheckWrapper', 'CommandConverterConfigurerWrapper', 'CommandWrapper', )

from ...backend.utils import copy_docs
from ...backend.export import export

from .content_parser import get_detail_for_value, ConverterFlag
from .cooldown import CooldownHandler

class CommandWrapper:
    """
    Command wrapper to add additional functionality to a command after it is created.
    
    Attributes
    ----------
    _wrapped : `Any`
        The wrapped object.
    """
    __slots__ = ('_wrapped',)
    def __new__(cls):
        """
        Creates a partial function to wrap a command.
        
        Subclasses should overwrite this method.
        """
        self = object.__new__(cls)
        self._wrapped = None
        return self
    
    def __call__(self, wrapped):
        """
        Wraps the given command.
        
        Parameters
        ----------
        wrapped : `Any`
            The slash command or other wrapper to wrap.
        
        Returns
        -------
        self : ``CommandWrapper``
        
        Raises
        ------
        RuntimeError
            The wrapper already wrapped something.
        """
        if (self._wrapped is not None):
            raise RuntimeError('The wrapper already wrapped something.')
        
        self._wrapped = wrapped
        return self
    
    def apply(self, command):
        """
        Applies the wrapper's changes on the respective command.
        
        Parameters
        ----------
        command : ``Command``
        """
        pass
    
    def __repr__(self):
        """Returns the command wrapper's representation."""
        return f'<{self.__class__.__name__} wrapped={self._wrapped!r}>'
    
    def fetch_function_and_wrappers_back(self):
        """
        Fetches back the source function and all the wrappers, the returns them.
        
        Returns
        -------
        function : `Any`
            The wrapped function.
        wrappers : `list` of ``CommandWrapper`` instances
            The fetched back wrappers.
        """
        wrappers = [self]
        maybe_wrapper = self._wrapped
        while True:
            if isinstance(maybe_wrapper, CommandWrapper):
                wrappers.append(maybe_wrapper)
                maybe_wrapper = maybe_wrapper._wrapped
            else:
                function = maybe_wrapper
                break
        
        wrappers.reverse()
        return function, wrappers


class CommandConverterConfigurerWrapper(CommandWrapper):
    """
    Command wrapper to modify parameter parsing settings for a specific command.
    
    Attributes
    ----------
    _wrapped : `Any`
        The wrapped object.
    _detail : ``ContentParserParameterDetail``
        Parsing detail example to match and modify of the source command.
    _flags : `None` or ``ConverterFlag``
        New flag to set to the parameter detail if any.
    _modifiers : `None` or `dict` of (`str`, `bool`) items
        Converter flag modifiers.
    """
    __slots__ = ('_detail', '_flags', '_modifiers', )
    
    def __new__(cls, annotation, flags=None, **modifiers):
        """
        Creates a new command wrapper instance for modifying converter flags.
    
        Raises
        ------
        TypeError
            - Modifier values can be `bool` instances.
            - If `annotation` was not given neither as `type` non as `bool` instance.
        ValueError
            - Invalid modifier name.
            - If `annotation`'s type is correct, but there is specified converter for it.
        """
        detail = get_detail_for_value(annotation)
        if (detail is None):
            raise TypeError(f'`annotation` can be given either as `type` or `str` instance, got '
                f'{annotation.__class__.__name__}.')
        
        if (flags is not None) and not isinstance(flags, ConverterFlag):
            if isinstance(flags, int):
                flags = ConverterFlag(flags)
            else:
                raise TypeError(f'`flag` can be given as `None`, `{ConverterFlag.__name__}` or as other `int` '
                    f'instance, got {flags.__class__.__name__}.')
        
        if modifiers:
            for key in modifiers.keys():
                if key not in ConverterFlag.__keys__:
                    raise ValueError(f'Invalid modifier: `{key!r}.')
            
            for key, value in modifiers.items():
                if not isinstance(value, bool):
                    raise TypeError(f'Modifier value can only be given as `bool` instance, got {key!r}={value!r}.')
            
            if (flags is not None):
                flags = flags.update_by_keys(**modifiers)
                modifiers = None
        else:
            modifiers = None
        
        self = object.__new__(cls)
        self._detail = detail
        self._flags = flags
        self._modifiers = modifiers
        self._wrapped = None
        return self
    
    @copy_docs(CommandWrapper.apply)
    def apply(self, command):
        if (self._flags is None) and (self._modifiers is None):
            return
        
        detail_example = self._detail
        command_function = command._command_function
        if (command_function is None):
            return
            
        content_parser = command_function._content_parser
        for parameter in content_parser._parameters:
            for detail in parameter._iter_details():
                if (detail.type is not detail_example.type):
                    continue
                
                if (detail.converter_setting is detail_example.converter_setting):
                    continue
                
                new_flags = self._flags
                if (new_flags is None):
                    new_flags = detail.flags.update_by_keys(**self._modifiers)
                
                detail.flags = new_flags
                continue


@export
class CommandCheckWrapper(CommandWrapper):
    """
    Command wrapper for checks.
    
    Attributes
    ----------
    _wrapped : `Any`
        The wrapped object.
    _check : ``CheckBase`` instance
        Check to add to the respective command.
    """
    __slots__ = ('_check',)
    def __new__(cls, check_type, *args, **kwargs):
        """
        Creates a partial function to wrap a command.
        
        Parameters
        ----------
        check_type : ``CheckBase`` subclass
            The check's type.
        *args : Parameters,
            Additional parameters to pass to the `check_type`'s constructor.
        **kwargs : Keyword parameters
            Additional parameters to pass to the `check_type`'s constructor.
        """
        check = check_type(*args, **kwargs)
        
        self = object.__new__(cls)
        self._check = check
        self._wrapped = None
        return self
    
    @copy_docs(CommandWrapper.apply)
    def apply(self, command):
        command.add_check(self._check)
    
    def __repr__(self):
        """Returns the command wrapper's representation."""
        return f'<{self.__class__.__name__} wrapped={self._wrapped!r} check={self._check!r}>'
    
    def __invert__(self):
        """Inverts the condition of the wrapped checks and returns a new command check wrapper."""
        new = object.__new__(type(self))
        new._check = ~self._check
        new._wrapped = None
        return new
    
    def __or__(self, other):
        """Connects the condition of two wrapped in `or` relation and returns a new command check wrapper."""
        if not isinstance(other, type(self)):
            return NotImplemented
        
        new = object.__new__(type(self))
        new._check = self._check|other._check
        new._wrapped = None
        return new
    
    def __and__(self, other):
        """Connects the condition of two wrapped in `and` relation and returns a new command check wrapper."""
        if not isinstance(other, type(self)):
            return NotImplemented
        
        new = object.__new__(type(self))
        new._check = self._check&other._check
        new._wrapped = None
        return new


class CommandCooldownWrapper(CommandWrapper):
    """
    Command wrapper for checks.
    
    Attributes
    ----------
    _wrapped : `Any`
        The wrapped object.
    _cooldown_handler : ``CooldownHandler``
        Cooldown to add to the respective command.
    """
    __slots__ = ('_cooldown_handler', )
    
    def __new__(cls, *args, **kwargs):
        """
        Creates a partial function to wrap a command.
        
        Parameters
        ----------
        *args : Parameters
            Parameters to pass to the `CooldownHandler`'s constructor.
        **kwargs : Keyword parameters
            Parameters to pass to the `CooldownHandler`'s constructor.
        
        Other Parameters
        ----------------
        for_ : `str`
            By what type of entity the cooldown should limit the command.
            
            Possible values:
             - `'user'`
             - `'channel'`
             - `'guild'`
         
        reset : `float`
            The reset time of the cooldown.
        
        limit : `int`
            The amount of calls after the respective command goes on cooldown.
        
        weight : `int`, Optional
            The weight of one call. Defaults to `1`.
        
        Raises
        ------
        TypeError
            - If `str` is not given as `str` instance.
            - If `weight` is not numeric convertable to `int`.
            - If `reset` is not numeric convertable to `float`.
            - If `limit` is not numeric convertable to `int`.
        ValueError
            - If `for_` is not given as any of the expected value.
        """
        cooldown_handler = CooldownHandler(*args, **kwargs)
        
        self = object.__new__(cls)
        self._cooldown_handler = cooldown_handler
        self._wrapped = None
        
        return self
    
    
    @copy_docs(CommandWrapper.apply)
    def apply(self, command):
        command.add_wrapper(self._cooldown_handler)



__all__ = ()

from ...discord.client import Client


UNLOADING_BEHAVIOUR_DELETE = 0
UNLOADING_BEHAVIOUR_KEEP = 1
UNLOADING_BEHAVIOUR_INHERIT = 2


SYNC_ID_GLOBAL = 0
SYNC_ID_MAIN = 1
SYNC_ID_NON_GLOBAL = 2


def raw_name_to_display(raw_name):
    """
    Converts the given raw application command name to it's display name.
    
    Parameters
    ----------
    raw_name : `str`
        The name to convert.
    
    Returns
    -------
    display_name : `str`
        The converted name.
    """
    return '-'.join([w for w in raw_name.strip('_ ').lower().replace(' ', '-').replace('_', '-').split('-') if w])


def normalize_description(description):
    """
    Normalizes a docstrings.
    
    Parameters
    ----------
    description : `None` or `str`
        The docstring to clear.
    
    Returns
    -------
    cleared : `None` or `str`
        The cleared docstring. If `docstring` was given as `None` or is detected as empty, will return `None`.
    """
    if description is None:
        return None
    
    lines = description.splitlines()
    for index in reversed(range(len(lines))):
        line = lines[index]
        line = line.strip()
        if line:
            lines[index] = line
        else:
            del lines[index]
    
    if not lines:
        return None
    
    return ' '.join(lines)


def _check_maybe_route(variable_name, variable_value, route_to, validator):
    """
    Helper class of ``SlasherApplicationCommand`` parameter routing.
    
    Parameters
    ----------
    variable_name : `str`
        The name of the respective variable
    variable_value : `str`
        The respective value to route maybe.
    route_to : `int`
        The value how much times the routing should happen. by default should be given as `0` if no routing was
        done yet.
    validator : `callable` or `None`
        A callable, what validates the given `variable_value`'s value and converts it as well if applicable.
    
    Returns
    -------
    processed_value : `str`
        Processed value returned by the `validator`. If routing is happening, then a `tuple` of those values is
        returned.
    route_to : `int`
        The amount of values to route to.
    
    Raises
    ------
    ValueError
        Value is routed but to a bad count amount.
    BaseException
        Any exception raised by `validator`.
    """
    if (variable_value is not None) and isinstance(variable_value, tuple):
        route_count = len(variable_value)
        if route_count == 0:
            processed_value = None
        elif route_count == 1:
            variable_value = variable_value[0]
            if variable_value is ...:
                variable_value = None
            
            if validator is None:
                processed_value = variable_value
            else:
                processed_value = validator(variable_value)
        else:
            if route_to == 0:
                route_to = route_count
            elif route_to == route_count:
                pass
            else:
                raise ValueError(f'`{variable_name}` is routed to `{route_count}`, meanwhile something else is '
                    f'already routed to `{route_to}`.')
            
            if validator is None:
                processed_value = variable_value
            else:
                processed_values = []
                for value in variable_value:
                    if (value is not ...):
                        value = validator(value)
                    
                    processed_values.append(value)
                
                processed_value = tuple(processed_values)
    
    else:
        if validator is None:
            processed_value = variable_value
        else:
            processed_value = validator(variable_value)
    
    return processed_value, route_to


RUNTIME_SYNC_HOOKS = []

def runtime_sync_hook_is_client_running(client):
    """
    Runtime sync hook to check whether a slash command should be registered and synced instantly when added or removed.
    
    Parameters
    ----------
    client : ``Client``
        The respective client of the ``Slasher``.
    """
    return client.running

RUNTIME_SYNC_HOOKS.append(runtime_sync_hook_is_client_running)





__all__ = ('User', 'ZEROUSER')

from ...env import CACHE_USER, CACHE_PRESENCE

from ...backend.utils import set_docs
from ...backend.export import include

from ..core import USERS

from ..preconverters import preconvert_snowflake, preconvert_str, preconvert_bool, preconvert_discriminator, \
    preconvert_flag, preconvert_color
from .preinstanced import Status
from .guild_profile import GuildProfile
from .client_user_base import ClientUserPBase, ClientUserBase
from .flags import UserFlag

create_partial_role_from_id = include('create_partial_role_from_id')
create_partial_user_from_id = include('create_partial_user_from_id')

if CACHE_PRESENCE:
    USER_BASE_CLASS = ClientUserPBase
else:
    USER_BASE_CLASS = ClientUserBase


class User(USER_BASE_CLASS):
    """
    Represents a Discord user.
    
    Attributes
    ----------
    id : `int`
        The user's unique identifier number.
    name : str
        The user's name.
    discriminator : `int`
        The user's discriminator. Given to avoid overlapping names.
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The user's avatar's type.
    banner_color : `None` or ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    guild_profiles : `dict` of (`int`, ``GuildProfile``) items
        A dictionary, which contains the user's guild profiles. If a user is member of a guild, then it should
        have a respective guild profile accordingly.
    is_bot : `bool`
        Whether the user is a bot or a user account.
    flags : ``UserFlag``
        The user's user flags.
    partial : `bool`
        Partial users have only their `.id` set and every other field might not reflect the reality.
    thread_profiles : `None` or `dict` (``ChannelThread``, ``ThreadProfile``) items
        A Dictionary which contains the thread profiles for the user in thread channel - thread profile relation.
        Defaults to `None`.
    activities : `None` or `list` of ``ActivityBase`` instances
        A list of the client's activities. Defaults to `None`
        
        > Only available if presence caching is enabled.
    status : ``Status``
        The user's display status.
        
        > Only available if presence caching is enabled.
    statuses : `dict` of (`str`, `str`) items
        The user's statuses for each platform.
        
        > Only available if presence caching is enabled.
    
    Notes
    -----
    Instances of this class are weakreferable.
    """
    __slots__ = ()
    
    if CACHE_PRESENCE:
        def __new__(cls, data, guild=None):
            try:
                user_data = data['user']
            except KeyError:
                user_data = data
                guild_profile_data = data.get('member', None)
            else:
                guild_profile_data = data
            
            user_id = int(user_data['id'])
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.thread_profiles = None
                self.status = Status.offline
                self.statuses = {}
                self.activities = None
                update = True
                
                USERS[user_id] = self
            else:
                update = self.partial
            
            if update:
                self.is_bot = user_data.get('bot', False)
                self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and (guild is not None):
                try:
                    profile = self.guild_profiles[guild.id]
                except KeyError:
                    guild.users[user_id] = self
                    self.guild_profiles[guild.id] = GuildProfile(guild_profile_data)
                else:
                    profile._set_joined(guild_profile_data)
            
            return self
    
    elif CACHE_USER:
        def __new__(cls, data, guild=None):
            try:
                user_data = data['user']
                guild_profile_data = data
            except KeyError:
                user_data = data
                guild_profile_data = data.get('member', None)
                
            user_id = int(user_data['id'])

            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.thread_profiles = None
                update = True
                
                USERS[user_id] = self
            else:
                update = self.partial
            
            if update:
                self.is_bot = user_data.get('bot', False)
                self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and (guild is not None):
                try:
                    guild_profile = self.guild_profiles[guild.id]
                except KeyError:
                    guild.users[user_id] = self
                    self.guild_profiles[guild.id] = GuildProfile(guild_profile_data)
                else:
                    guild_profile._set_joined(guild_profile_data)
                    
            return self
    
    else:
        def __new__(cls, data, guild=None):
            try:
                user_data = data['user']
                guild_profile_data = data
            except KeyError:
                user_data = data
                guild_profile_data = data.get('member', None)
            
            user_id = int(user_data['id'])
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.thread_profiles = None
                
                USERS[user_id] = self
            
            self.is_bot = user_data.get('bot', False)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and (guild is not None):
                self.guild_profiles[guild.id] = GuildProfile(guild_profile_data)
            
            return self
    
    set_docs(__new__,
        """
        First tries to find the user by id. If fails, then creates a new ``User`` object. If guild was given
        and the given data contains member data as well, then it will create a respective guild profile for the user
        too.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received user data.
        guild : ``Guild`` or `None`, Optional
            A respective guild from where the user data was received. It is picked up if the given data includes
            guild member data as well.
        
        Returns
        -------
        user : ``ClientUserBase``
        """)
    
    
    @classmethod
    def precreate(cls, user_id, **kwargs):
        """
        Precreates a user by creating a partial one with the given parameters. When the user is loaded, the precreated
        one is picked up and is updated. If an already existing user would be precreated, returns that instead of
        creating a new one, and updates it only, if it is still a partial one.
        
        Parameters
        ----------
        user_id : `int` or `str`
            The user's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the user.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The user's ``.name``.
        discriminator : `int` or `str` instance, Optional (Keyword only)
            The user's ``.discriminator``. Is accepted as `str` instance as well and will be converted to `int`.
        
        avatar : `None`, ``Icon`` or `str`, Optional (Keyword only)
            The user's avatar.
            
            > Mutually exclusive with `avatar_type` and `avatar_hash`.
        
        avatar_type : ``IconType``, Optional (Keyword only)
            The user's avatar's type.
            
            > Mutually exclusive with `avatar_type`.
        
        avatar_hash : `int`, Optional (Keyword only)
            The user's avatar's hash.
            
            > Mutually exclusive with `avatar`.
        
        banner : `None`, ``Icon`` or `str`, Optional (Keyword only)
            The user's banner.
            
            > Mutually exclusive with `banner_type` and `banner_hash`.
        
        banner_color : `None` or ``Color``
            The user's banner color.
        
        banner_type : ``IconType``, Optional (Keyword only)
            The user's banner's type.
            
            > Mutually exclusive with `banner_type`.
        
        banner_hash : `int`, Optional (Keyword only)
            The user's banner hash.
            
            > Mutually exclusive with `banner`.
        
        flags : ``UserFlag`` or `int` instance, Optional (Keyword only)
            The user's ``.flags``. If not passed as ``UserFlag``, then will be converted to it.
        
        Returns
        -------
        user : ``ClientUserBase``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        user_id = preconvert_snowflake(user_id, 'user_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 32)
                processable.append(('name', name))
            
            try:
                discriminator = kwargs.pop('discriminator')
            except KeyError:
                pass
            else:
                discriminator = preconvert_discriminator(discriminator)
                processable.append(('discriminator', discriminator))
            
            cls.avatar.preconvert(kwargs, processable)
            cls.banner.preconvert(kwargs, processable)
            
            try:
                is_bot = kwargs.pop('is_bot')
            except KeyError:
                pass
            else:
                is_bot = preconvert_bool(is_bot, 'is_bot')
                processable.append(('is_bot', is_bot))
            
            try:
                flags = kwargs.pop('flags')
            except KeyError:
                pass
            else:
                flags = preconvert_flag(flags, 'flags', UserFlag)
                processable.append(('flags', flags))
            
            try:
                banner_color = kwargs.pop('banner_color')
            except KeyError:
                pass
            else:
                banner_color = preconvert_color(banner_color, 'banner_color', True)
                processable.append(('banner_color', banner_color))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}.')
        
        else:
            processable = None
        
        self = create_partial_user_from_id(user_id)
        if not self.partial:
            return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
    
    
    if CACHE_PRESENCE:
        @classmethod
        def _create_and_update(cls, data, guild=None):
            try:
                user_data = data['user']
                guild_profile_data = data
            except KeyError:
                user_data = data
                guild_profile_data = None
            
            user_id = int(user_data['id'])
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                self.status = Status.offline
                self.statuses = {}
                self.activities = None
                
                USERS[user_id] = self
            
            self.is_bot = user_data.get('bot', False)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and (guild is not None):
                try:
                    guild_profile = self.guild_profiles[guild.id]
                except KeyError:
                    guild.users[user_id] = self
                    self.guild_profiles[guild.id] = GuildProfile(guild_profile_data)
                else:
                    guild_profile._set_joined(guild_profile_data)
                    guild_profile._update_attributes(guild_profile_data)
            
            return self
        
    elif CACHE_USER:
        @classmethod
        def _create_and_update(cls, data, guild=None):
            try:
                user_data = data['user']
                guild_profile_data = data
            except KeyError:
                user_data = data
                guild_profile_data = None
            
            user_id = int(user_data['id'])
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                
                USERS[user_id] = self
            
            self.is_bot = user_data.get('bot', False)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and (guild is not None):
                try:
                    guild_profile = self.guild_profiles[guild.id]
                except KeyError:
                    guild.users[user_id] = self
                    self.guild_profiles[guild.id] = GuildProfile(guild_profile_data)
                else:
                    guild_profile._set_joined(guild_profile_data)
                    guild_profile._update_attributes(guild_profile_data)
            
            return self
        
    else:
        @classmethod
        def _create_and_update(cls, data, guild=None):
            try:
                user_data = data['user']
            except KeyError:
                user_data = data
                guild_profile_data = None
            else:
                guild_profile_data = data
            
            user_id = int(user_data['id'])
            
            try:
                self = USERS[user_id]
            except KeyError:
                self = object.__new__(cls)
                self.id = user_id
                self.guild_profiles = {}
                
                USERS[user_id] = self
            
            self.is_bot = user_data.get('bot', False)
            self._update_attributes(user_data)
            
            if (guild_profile_data is not None) and (guild is not None):
                self.guild_profiles[guild.id] = GuildProfile(guild_profile_data)
            
            return self
    
    set_docs(_create_and_update,
        """
        Creates a user with the given data. If the user already exists, updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received user data.
        guild : ``Guild`` or `None`, Optional
            A respective guild from where the user data was received. Picked up if the given data includes
            guild member data as well.
        
        Returns
        -------
        user : ``ClientUserBase``
        """)


ZEROUSER = User._create_empty(0)



__all__ = ('EXTRA_EMBED_TYPES', 'EmbedAuthor', 'EmbedBase', 'EmbedField', 'EmbedFooter', 'EmbedImage', 'EmbedProvider',
    'EmbedThumbnail', 'EmbedVideo', )

from datetime import datetime

from ..utils import url_cutter
from ..color import Color

EXTRA_EMBED_TYPES = frozenset(('application_news', 'article', 'gifv', 'image', 'link', 'tweet', 'video'))


class EmbedThumbnail:
    """
    Represents an embed's thumbnail.
    
    Attributes
    ----------
    height : `int`
        The height of the thumbnail. Defaults to `0`.
    proxy_url : `None` or `str`
        A proxied url of the thumbnail.
    url : `None` or `str`
        The url of the thumbnail.
    width : `int`
        The width of the thumbnail. Defaults to `0`.
    """
    __slots__ = ('height', 'proxy_url', 'url', 'width',)
    
    def __init__(self, url):
        """
        Creates an embed thumbnail with the given `url`.
        
        Parameters
        ----------
        url : `str`
            The url of the thumbnail. Can be http(s) or attachment.
        """
        self.url = url
        self.proxy_url = None
        self.height = 0
        self.width = 0
    
    
    def __len__(self):
        """Returns the embed thumbnail's contents' length."""
        return 0
    
    
    def __bool__(self):
        """Returns whether the embed thumbnail is not empty."""
        return (self.url is not None)
    
    
    def __repr__(self):
        """Returns the representation of the embed thumbnail."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' url=',
        ]
        
        url = self.url
        if url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            url = url_cutter(url)
            repr_parts.append(url)
            repr_parts.append('\'')
        
        repr_parts.append(', size=')
        repr_parts.append(str(self.width))
        repr_parts.append('x')
        repr_parts.append(str(self.height))
    
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embed thumbnails are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return (self.url == other.url)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a embed thumbnail object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed thumbnail data received from Discord.
        
        Returns
        -------
        embed_thumbnail : ``EmbedThumbnail``
        """
        self = object.__new__(cls)
        self.url = data.get('url', None)
        self.proxy_url = data.get('proxy_url', None)
        self.height = data.get('height', 0)
        self.width = data.get('width', 0)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed thumbnail to json serializable `dict` representing it.
        
        Returns
        -------
        thumbnail_data : `dict` of (`str`, `Any`) items
        """
        return {
            'url' : self.url,
        }
    
    def copy(self):
        """
        Copies the ``EmbedThumbnail`` returning a new one.
        
        Returns
        -------
        new : ``EmbedThumbnail``
        """
        new = object.__new__(type(self))
        
        new.url = self.url
        new.proxy_url = None
        new.height = 0
        new.width = 0
        
        return self
    
    
    def copy_with(self, *, url=...):
        """
        Copies the ``EmbedThumbnail`` and updates it with the given parameters.
        
        Parameters
        ----------
        url : `str`, Optional (Keyword only)
            The url of the thumbnail. Can be http(s) or attachment.
        
        Returns
        -------
        new : ``EmbedThumbnail``
        """
        if url is ...:
            url = self.url
        
        new = object.__new__(type(self))
        new.url = url
        new.proxy_url = None
        new.height = 0
        new.width = 0
        
        return self


class EmbedVideo:
    """
    Represents an embed's video.
    
    Embed videos cannot be sent, they are receive only.
    
    Attributes
    ----------
    height : `int`
        The height of the video. Defaults to `0`.
    proxy_url : `None` or `str`
        A proxied url of the video.
    url : `None` or `str`
        The url of the video.
    width : `int`
        The width of the video. Defaults to `0`.
    """
    __slots__ = ('height', 'proxy_url', 'url', 'width',)
    
    def __init__(self):
        """
        Creates a new embed video object. Because embed videos cannot be sent, it accepts no parameters and just creates
        an empty embed video object with default attributes.
        """
        self.url = None
        self.proxy_url = None
        self.height = 0
        self.width = 0
    
    
    def __len__(self):
        """Returns the embed video's contents' length."""
        return 0
    
    
    def __bool__(self):
        """Returns whether the embed video is not empty."""
        return (self.url is not None)
    
    
    def __repr__(self):
        """Returns the representation of the embed video."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' url=',
        ]
        
        url = self.url
        if url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            url = url_cutter(url)
            repr_parts.append(url)
            repr_parts.append('\'')
        
        repr_parts.append(', size=')
        repr_parts.append(str(self.width))
        repr_parts.append('x')
        repr_parts.append(str(self.height))
    
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embed videos are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return (self.url == other.url)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed video object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed video data received from Discord.
        
        Returns
        -------
        embed_video : ``EmbedVideo``
        """
        self = object.__new__(cls)
        self.url = data.get('url', None)
        self.proxy_url = data.get('proxy_url', None)
        self.height = data.get('height', 0)
        self.width = data.get('width', 0)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed video to json serializable `dict` representing it.
        
        Because embed videos cannot be sent, it always returns an empty `dict`.
        
        Returns
        -------
        video_data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def copy(self):
        """
        Copies the ``EmbedVideo`` returning a new one.
        
        > Since embed videos are receive only, the method will return the instance itself.
        
        Returns
        -------
        new : ``EmbedVideo``
        """
        return self
    
    
    def copy_with(self):
        """
        Copies the ``EmbedVideo`` and updates it with the given parameters.
        
        > Since embed videos are receive only, the method will return the instance itself.
        
        Returns
        -------
        new : ``EmbedVideo``
        """
        return self


class EmbedImage:
    """
    Represents an embed's image.
    
    Attributes
    ----------
    height : `int`
        The height of the image. Defaults to `0`.
    proxy_url : `None` or `str`
        A proxied url of the image.
    url : `None` or `str`
        The url of the image.
    width : `int`
        The width of the image. Defaults to `0`.
    """
    __slots__ = ('height', 'proxy_url', 'url', 'width',)
    
    def __init__(self, url):
        """
        Creates an embed image with the given `url`.
        
        Parameters
        ----------
        url : `str`
            The url of the image. Can be http(s) or attachment.
        """
        self.url = url
        self.proxy_url = None
        self.height = 0
        self.width = 0
    
    
    def __len__(self):
        """Returns the embed image's contents' length."""
        return 0
    
    
    def __bool__(self):
        """Returns whether the embed image is not empty."""
        return (self.url is not None)
    
    
    def __repr__(self):
        """Returns the representation of the embed image."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' url=',
        ]
        
        url = self.url
        if url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            url = url_cutter(url)
            repr_parts.append(url)
            repr_parts.append('\'')
            
        repr_parts.append(', size=')
        repr_parts.append(str(self.width))
        repr_parts.append('x')
        repr_parts.append(str(self.height))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embed images are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return (self.url == other.url)
    
    
    @classmethod
    def from_data(cls,data):
        """
        Creates an embed image object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed image data received from Discord.
        
        Returns
        -------
        embed_image : ``EmbedImage``
        """
        self = object.__new__(cls)
        self.url = data.get('url', None)
        self.proxy_url = data.get('proxy_url', None)
        self.height = data.get('height', 0)
        self.width = data.get('width', 0)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed image to json serializable `dict` representing it.
        
        Returns
        -------
        image_data : `dict` of (`str`, `Any`) items
        """
        return {
            'url' : self.url,
        }
    

    def copy(self):
        """
        Copies the ``EmbedImage`` returning a new one.
        
        Returns
        -------
        new : ``EmbedImage``
        """
        new = object.__new__(type(self))
        
        new.url = self.url
        new.proxy_url = None
        new.height = 0
        new.width = 0
        
        return self
    
    
    def copy_with(self, *, url=...):
        """
        Copies the ``EmbedImage`` and updates it with the given parameters.
        
        Parameters
        ----------
        url : `str`, Optional (Keyword only)
            The url of the image. Can be http(s) or attachment.
        
        Returns
        -------
        new : ``EmbedImage``
        """
        if url is ...:
            url = self.url
        
        new = object.__new__(type(self))
        new.url = url
        new.proxy_url = None
        new.height = 0
        new.width = 0
        
        return self


class EmbedProvider:
    """
    Represents an embed's provider.
    
    Embed providers cannot be sent, they are receive only.
    
    Attributes
    ----------
    name : `None` or `str`
        The name of the provider.
    url : `None` or `str`
        The url of the provider.
    """
    __slots__ = ('name', 'url',)
    
    def __init__(self):
        self.name = None
        self.url = None
    
    
    def __len__(self):
        """Returns the embed provider's contents' length."""
        name = self.name
        if name is None:
            return 0
        
        return len(name)
    
    
    def __bool__(self):
        """Returns whether the embed provider is not empty."""
        if (self.url is not None):
            return True
        
        name = self.name
        if (name is not None) and name:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the representation of the embed provider."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' length=',
            str(len(self)),
            ', url='
        ]
        
        url = self.url
        if url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            url = url_cutter(url)
            repr_parts.append(url)
            repr_parts.append('\'')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embed providers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.name != other.name:
            return False
        
        if self.url != other.url:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new embed provider object. Because embed providers cannot be sent, it accepts no parameters and just
        creates an empty embed provider object with default attributes.
        
        Returns
        -------
        self : ``EmbedProvider``
        """
        self = object.__new__(cls)
        self.name = data.get('name', None)
        self.url = data.get('url', None)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed provider to json serializable `dict` representing it.
        
        Because embed provider cannot be sent, it always returns an empty `dict`.
        
        Returns
        -------
        provider_data : `dict` of (`str`, `Any`) items
        """
        return {}


    def copy(self):
        """
        Copies the ``EmbedProvider`` returning a new one.
        
        > Since embed providers are receive only, the method will return the instance itself.
        
        Returns
        -------
        new : ``EmbedProvider``
        """
        return self
    
    
    def copy_with(self):
        """
        Copies the ``EmbedProvider`` and updates it with the given parameters.
        
        > Since embed providers are receive only, the method will return the instance itself.
        
        Returns
        -------
        new : ``EmbedProvider``
        """
        return self


class EmbedAuthor:
    """
    Represents an embed's author.
    
    Attributes
    ----------
    icon_url : `None` or `str`
        Url of the author's icon.
    name : `None` or `str`
        The name of the author.
    proxy_icon_url : `None` or `str`
        A proxied url to the url of the author's icon.
    url : `None` or `str`
        The url of the author.
    """
    __slots__ = ('icon_url', 'name', 'proxy_icon_url', 'url',)
    def __init__(self, icon_url=None, name=None, url=None):
        """
        Creates an embed author with the given parameters.
        
        Parameters
        ----------
        icon_url : `str`, Optional
            An url of the author's icon. Can be http(s) or attachment.
        name : `str`, Optional
            The name of the author.
        url : `str`, Optional
            The url of the author.
        """
        self.icon_url = icon_url
        self.name = name
        self.url = url
        self.proxy_icon_url = None
    
    
    def __len__(self):
        """Returns the embed author's contents' length."""
        name = self.name
        if name is None:
            return 0
        
        return len(name)
    
    
    def __bool__(self):
        """Returns whether the embed author is not empty."""
        name = self.name
        if (name is not None) and name:
            return True
        
        if (self.url is not None):
            return True
        
        if (self.icon_url is not None):
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the representation of the embed author."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' length=',
            str(len(self)),
            ', url='
        ]
        
        url = self.url
        if url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            url = url_cutter(url)
            repr_parts.append(url)
            repr_parts.append('\'')
        
        repr_parts.append(', icon_url=')
        icon_url = self.icon_url
        if icon_url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            icon_url = url_cutter(icon_url)
            repr_parts.append(icon_url)
            repr_parts.append('\'')
            
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def __eq__(self, other):
        """Returns whether the two embed authors are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.icon_url != other.icon_url:
            return False
        
        if self.name != other.name:
            return False
        
        if self.url != other.url:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed author object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed author data received from Discord.
        
        Returns
        -------
        embed_author : ``EmbedAuthor``
        """
        self = object.__new__(cls)
        self.name = data.get('name', None)
        self.url = data.get('url', None)
        self.icon_url = data.get('icon_url', None)
        self.proxy_icon_url = data.get('proxy_icon_url', None)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed author to json serializable `dict` representing it.
        
        Returns
        -------
        author_data : `dict` of (`str`, `Any`) items
        """
        author_data = {}
    
        name = self.name
        if (name is not None):
            author_data['name'] = name
        
        url = self.url
        if (url is not None):
            author_data['url'] = url
        
        icon_url = self.icon_url
        if (icon_url is not None):
            author_data['icon_url'] = icon_url
        
        return author_data
    
    
    def copy(self):
        """
        Copies the ``EmbedAuthor`` returning a new one.
        
        Returns
        -------
        new : ``EmbedAuthor``
        """
        new = object.__new__(type(self))
        
        new.icon_url = self.icon_url
        new.name = self.name
        new.url = self.url
        new.proxy_icon_url = None
        
        return self
    
    
    def copy_with(self, *, icon_url=..., name=..., url=...):
        """
        Copies the ``EmbedAuthor`` and updates it with the given parameters.
        
        Parameters
        ----------
        icon_url : `str`, Optional (Keyword only)
            An url of the author's icon. Can be http(s) or attachment.
        name : `str`, Optional (Keyword only)
            The name of the author.
        url : `str`, Optional (Keyword only)
            The url of the author.
        
        Returns
        -------
        new : ``EmbedAuthor``
        """
        if icon_url is ...:
            icon_url = self.icon_url
        
        if name is ...:
            name = self.name
        
        if url is ...:
            url = self.url
        
        new = object.__new__(type(self))
        
        new.icon_url = icon_url
        new.name = name
        new.url = url
        new.proxy_icon_url = None
        
        return self
    

class EmbedFooter:
    """
    Represents an embed's footer.
    
    Attributes
    ----------
    icon_url :`None` or `str`
        Url of the embed footer's icon.
    proxy_icon_url : `None` or `str`
        A proxied url of the embed footer's icon.
    text : `str`
        The embed footer's text.
    """
    __slots__ = ('icon_url', 'proxy_icon_url', 'text',)
    
    
    def __init__(self, text, icon_url=None):
        """
        Creates an embed footer with the given parameters.
        
        Parameters
        ----------
        text : `str`
            The footer's text.
        icon_url : `str`, Optional
            An url of the footer's icon. Can be http(s) or attachment.
        """
        self.text = text
        self.icon_url = icon_url
        self.proxy_icon_url = None
    
    
    def __len__(self):
        """Returns the embed footer's contents' length."""
        return len(self.text)
    
    
    def __bool__(self):
        """Returns whether the embed footer is not empty."""
        if self.text:
            return True
        
        if (self.icon_url is not None):
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the representation of the embed footer."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' length=',
            str(len(self)),
            ', url='
        ]
        
        icon_url = self.icon_url
        if icon_url is None:
            repr_parts.append('None')
        else:
            repr_parts.append('\'')
            icon_url = url_cutter(icon_url)
            repr_parts.append(icon_url)
            repr_parts.append('\'')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embed footers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.icon_url != other.icon_url:
            return False
        
        if self.text != other.text:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed footer object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed footer data received from Discord.
        
        Returns
        -------
        embed_footer : ``EmbedFooter``
        """
        self = object.__new__(cls)
        self.text = data['text']
        self.icon_url = data.get('icon_url', None)
        self.proxy_icon_url = data.get('proxy_icon_url', None)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed footer to json serializable `dict` representing it.
        
        Returns
        -------
        footer_data : `dict` of (`str`, `Any`) items
        """
        footer_data = {
            'text' : self.text,
        }
        
        icon_url = self.icon_url
        if (icon_url is not None):
            footer_data['icon_url'] = icon_url
            
        return footer_data
    
    
    def copy(self):
        """
        Copies the ``EmbedFooter`` returning a new one.
        
        Returns
        -------
        new : ``EmbedFooter``
        """
        new = object.__new__(type(self))
        
        new.text = self.text
        new.icon_url = self.icon_url
        new.proxy_icon_url = None
        
        return self
    
    
    def copy_with(self, *, text=..., icon_url=...):
        """
        Copies the ``EmbedFooter`` and updates it with the given parameters.
        
        Parameters
        ----------
        text : `str`, Optional (Keyword only)
            The footer's text.
        icon_url : `str`, Optional (Keyword only)
            An url of the footer's icon. Can be http(s) or attachment.
        
        Returns
        -------
        new : ``EmbedFooter``
        """
        if text is ...:
            text = self.text
        
        if icon_url is ...:
            icon_url = self.icon_url
        
        new = object.__new__(type(self))
        
        new.text = text
        new.icon_url = icon_url
        new.proxy_icon_url = None
        
        return self


class EmbedField:
    """
    Represents an embed's field.
    
    Attributes
    ----------
    inline : `bool`
        Whether this field should display inline. Defaults to `False`.
    name : `str`
        The name of the field.
    value : `str`
        The value of the field.
    """
    __slots__ = ('inline', 'name', 'value',)
    
    def __init__(self, name, value, inline=False):
        """
        Creates an embed field with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool`, Optional
            Whether this field should display inline.
        """
        self.name = name
        self.value = value
        self.inline = inline
    
    
    def __len__(self):
        """Returns the embed field's contents' length."""
        return len(self.name)+len(self.value)
    
    
    def __bool__(self):
        """Returns whether the embed field is not empty."""
        if self.name:
            return True
        
        if self.value:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the representation of the embed field."""
        return f'<{self.__class__.__name__} length={len(self)}, inline={self.inline}>'
    
    
    def __eq__(self, other):
        """Returns whether the two embed fields are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.name != other.name:
            return False
        
        if self.value != other.value:
            return False
        
        if self.inline != other.inline:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed field object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed field data received from Discord.
        
        Returns
        -------
        embed_field : ``EmbedField``
        """
        self = object.__new__(cls)
        
        self.name = data['name']
        self.value = data['value']
        self.inline = data.get('inline',False)
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed field to json serializable `dict` representing it.
        
        Returns
        -------
        field_data : `dict` of (`str`, `Any`) items
        """
        field_data = {
            'name': self.name,
            'value': self.value,
        }
        
        inline = self.inline
        if inline:
            field_data['inline'] = inline
        
        return field_data


    def copy(self):
        """
        Copies the ``EmbedField`` returning a new one.
        
        Returns
        -------
        new : ``EmbedField``
        """
        new = object.__new__(type(self))
        
        new.name = self.name
        new.value = self.value
        new.inline = self.inline
        
        return self
    
    
    def copy_with(self, *, name=..., value=..., inline=...):
        """
        Copies the ``EmbedField`` and updates it with the given parameters.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The name of the field.
        value : `str`, Optional (Keyword only)
            The value of the field.
        inline : `bool`, Optional (Keyword only)
            Whether this field should display inline.
        
        Returns
        -------
        new : ``EmbedField``
        """
        if name is ...:
            name = self.name
        
        if value is ...:
            value = self.value
        
        if inline is ...:
            inline = self.inline
        
        new = object.__new__(type(self))
        
        new.name = name
        new.value = value
        new.inline = inline
        
        return self


EMBED_ATTRIBUTES = frozenset((
    'author',
    'color',
    'description',
    'fields',
    'footer',
    'image',
    'provider',
    'thumbnail',
    'timestamp',
    'title',
    'type',
    'url',
    'video',
))

class EmbedBase:
    """
    Base class for Discord embedded contents. Should be taken as a guide for implementing custom embed classes.
    
    Abstract Attributes
    -------------------
    author : `None` or ``EmbedAuthor``
        Author information.
    color : `None`, ``Color`` or `int`
        The color code of the embed. Passing `0` means black, not like at the case of roles.
    description : `None` or `str`
        The main content of the embed.
    fields : `list` of ``EmbedField``
        Fields' information.
    footer : `None` or ``EmbedFooter``
        Footer information.
    image : `None` or ``EmbedImage``
        Image information.
    provider : `None` or ``EmbedProvider``
        Provider information.
    thumbnail : `None` or ``EmbedThumbnail``
        Thumbnail information.
    timestamp : `None` or `datetime`
        Timestamp of the embed's content. Shows up next to the ``.footer`` separated with a `'|'` character.
    title : `None` or `str`
        The title of the embed. Shows at the top with intense white characters.
    type : `None` or `str`
        The type of the embed. Can be one of `EXTRA_EMBED_TYPES`'s elements. Webhook embeds' type must be `'rich'`.
    url : `None` or `str`
        Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
    video : `None` or `EmbedVideo`
        Video information.
    """
    
    __slots__ = ()
    
    author : {None, EmbedAuthor}
    color : {None, Color, int}
    description : {None, str}
    fields : (list, EmbedField)
    footer : {None, EmbedFooter}
    image : {None, EmbedImage}
    provider : {None, EmbedProvider}
    thumbnail : {None, EmbedThumbnail}
    timestamp : {None, datetime}
    title : {None, str}
    type : {None, str}
    url : {None, str}
    video : {None, EmbedVideo}
    
    
    def __len__(self):
        """Returns the embed's contents' length."""
        result = 0
        
        title = self.title
        if (title is not None):
            result += len(title)
        
        description = self.description
        if (description is not None):
            result += len(description)
        
        title = self.title
        if (title is not None):
            result += len(title)
        
        footer = self.footer
        if (footer is not None):
            result += len(footer.text)
        
        author = self.author
        if (author is not None):
            name = author.name
            if (name is not None):
                result += len(name)
        
        for field in self.fields:
            result += len(field.name)
            result += len(field.value)
        
        return result
    
    
    def __bool__(self):
        """Returns whether the embed is not empty."""
        author = self.author
        if (author is not None) and author:
            return True
        
        if (self.color is not None):
            return True
        
        description = self.description
        if (description is not None) and description:
            return True
        
        if self.fields:
            return True
        
        footer = self.footer
        if (footer is not None) and footer:
            return True
        
        image = self.image
        if (image is not None) and image:
            return True
        
        provider = self.provider
        if (provider is not None) and provider:
            return True
        
        thumbnail = self.thumbnail
        if (thumbnail is not None) and thumbnail:
            return True
        
        if (self.timestamp is not None):
            return True
        
        title = self.title
        if (title is not None) and title:
            return True
        
        if (self.url is not None):
            return True
        
        video = self.video
        if (video is not None) and video:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the representation of the embed."""
        return f'<{self.__class__.__name__} length={len(self)}>'
    
    
    def __eq__(self, other):
        """Returns whether the two embeds are equal."""
        if not isinstance(other, EmbedBase):
            return NotImplemented
        
        if self.title != other.title:
            return False
            
        if self.description != other.description:
            return False
        
        if self.color != other.color:
            return False
        
        if self.url != other.url:
            return False
            
        if self.timestamp != other.timestamp:
            return False
            
        if self.type != other.type:
            return False
        
        if self.footer != other.footer:
            return False
            
        if self.image != other.image:
            return False
            
        if self.thumbnail != other.thumbnail:
            return False
        
        if self.video != other.video:
            return False
        
        if self.author != other.author:
            return False
        
        if self.fields != other.fields:
            return False
        
        return True
    
    
    def from_data(self, data):
        """
        Creates a new embed from the given data received from Discord.
        
        > Subclasses should overwrite this method.
        
        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError
    
    
    def to_data(self):
        """
        Converts the embed to json serializable `dict` representing it.
        
        > Subclasses should overwrite this method.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def clear(self):
        """
        Clears the embed.
        """
    
    
    @property
    def contents(self):
        """
        Returns the embed's contents.
        
        > Subclasses should overwrite this method.
        
        The embed's contents are the following:
        - `.title`
        - `.description`
        - `.author.name`
        - `.footer.text`
        - `.fields[n].name`
        - `.fields[n].value`
        
        Returns
        -------
        contents : `list` of `str`
        """
        return []
    
    
    def copy(self):
        """
        Copies the embed returning a new one.
        
        Returns
        -------
        new : ``EmbedBase``
        """
        return
    
    
    def copy_with(self, **kwargs):
        """
        Copies the embed and updates it with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
             Keyword parameters referencing attributes.
         
        Other Parameters
        ----------------
        author : `None` or ``EmbedAuthor``, Optional (Keyword only)
            Author information.
        color : `None`, ``Color`` or `int`, Optional (Keyword only)
            The color code of the embed. Passing `0` means black, not like at the case of roles.
        description : `None` or `str`, Optional (Keyword only)
            The main content of the embed.
        fields : `list` of ``EmbedField``, Optional (Keyword only)
            Fields' information.
        footer : `None` or ``EmbedFooter``, Optional (Keyword only)
            Footer information.
        image : `None` or ``EmbedImage``, Optional (Keyword only)
            Image information.
        provider : `None` or ``EmbedProvider``, Optional (Keyword only)
            Provider information.
        thumbnail : `None` or ``EmbedThumbnail``, Optional (Keyword only)
            Thumbnail information.
        timestamp : `None` or `datetime`, Optional (Keyword only)
            Timestamp of the embed's content. Shows up next to the ``.footer`` separated with a `'|'` character.
        title : `None` or `str`, Optional (Keyword only)
            The title of the embed. Shows at the top with intense white characters.
        type : `None` or `str`, Optional (Keyword only)
            The type of the embed. Can be one of `EXTRA_EMBED_TYPES`'s elements. Webhook embeds' type must be `'rich'`.
        url : `None` or `str`, Optional (Keyword only)
            Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
        video : `None` or `EmbedVideo`, Optional (Keyword only)
            Video information.
        
        Returns
        -------
        new : ``EmbedBase``
        """
        for key in kwargs:
            if key not in EMBED_ATTRIBUTES:
                raise TypeError(f'Unused or unsettable attribute: `{key}`')
        
        return



__all__ = ('AUTO_ARCHIVE_DEFAULT', 'AUTO_ARCHIVE_OPTIONS', 'ChannelThread', )

from ...backend.utils import copy_docs
from ...backend.export import export

from ..core import CHANNELS, GUILDS
from ..permission import Permission
from ..permission.permission import PERMISSION_NONE, PERMISSION_TEXT_DENY, PERMISSION_VOICE_DENY, \
    PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_MASK_MANAGE_MESSAGES, PERMISSION_MASK_SEND_MESSAGES, \
    PERMISSION_DENY_SEND_MESSAGES_ONLY
from ..user import ZEROUSER, create_partial_user_from_id
from ..user.thread_profile import thread_user_create
from ..preconverters import preconvert_snowflake, preconvert_str, preconvert_int, preconvert_int_options, \
    preconvert_bool
from ..utils import timestamp_to_datetime

from .channel_base import ChannelBase
from .channel_guild_base import ChannelGuildBase
from .channel_text_base import ChannelTextBase


AUTO_ARCHIVE_DEFAULT = 3600
AUTO_ARCHIVE_OPTIONS = frozenset((3600, 86400, 259200, 604800))

CHANNEL_THREAD_NAMES = {
     9: None,
    10: 'announcements',
    11: 'public',
    12: 'private',
}

@export
class ChannelThread(ChannelGuildBase, ChannelTextBase):
    """
    Represents a ``Guild`` thread channel
    
    Attributes
    ----------
    id : `int`
        The unique identifier of the channel.
    _permission_cache : `None` or `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent : `None` or ``ChannelText``
        The text channel from where the thread is created from.
    guild_id : `int`
        The channel's guild's identifier.
    name : `str`
        The channel's name.
    _message_history_collector :  `None` or ``MessageHistoryCollector``
        Collector for the channel's message history.
    _message_keep_limit : `int`
        The channel's own limit of how much messages it should keep before removing their reference.
    message_history_reached_end : `bool`
        Whether the channel's message's are loaded till their end. If the channel's message history reached it's end
        no requests will be requested to get older messages.
    messages : `deque` of ``Message`` objects
        The channel's message history.
    archived : `bool`
        Whether the thread s archived.
    archived_at : `None` or `datetime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    invitable : `bool`
        Whether non-moderators can invite other non-moderators to the threads. Only applicable for private threads.
    open : `bool`
        Whether the thread channel is open.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages` or `manage_channel` permissions are unaffected.
    thread_users : `None` or `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    type : `int` = `12`
        The channel's Discord side type.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `12`
        The preferred channel type, if there is no channel type included.
    INTERCHANGE : `tuple` of `int` = `(10, 11, 12,)`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    ORDER_GROUP : `int` = `9`
        An order group what defined which guild channel type comes after the other one.
    REPRESENTED_TYPES : `tuple` = (`10`, `11`, `12`,)
        The type values which ``ChannelThread`` might represent.
    """
    __slots__ = ('archived', 'archived_at', 'auto_archive_after', 'invitable', 'open', 'owner_id', 'slowmode',
        'thread_users', 'type')
    
    DEFAULT_TYPE = 12
    ORDER_GROUP = 9
    INTERCHANGE = ()
    REPRESENTED_TYPES = (10, 11, 12,)
    
    def __new__(cls, data, client, guild_id):
        """
        Creates a guild thread channel from the channel data received from Discord. If the channel already exists and
        if it is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        client : `None` or ``Client``
            The client, who received the channel's data, if any.
        guild_id : `int`
            The channel's guild's identifier.
        """
        channel_id = int(data['id'])
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = channel_id
            CHANNELS[channel_id] = self
            self._messageable_init()
            self.thread_users = None
            update = True
        else:
            if self.clients:
                update = False
            else:
                update = True
        
        if update:
            self._permission_cache = None
            self.type = data['type']
            self._init_parent(data, guild_id)
            self._update_attributes(data)
            
            owner_id = data.get('owner_id', None)
            if owner_id is None:
                owner_id = 0
            else:
                owner_id = int(owner_id)
            self.owner_id = owner_id
        
        if (client is not None):
            try:
                thread_user_data = data['member']
            except KeyError:
                pass
            else:
                thread_user_create(self, client, thread_user_data)
        
        return self
    
    
    @copy_docs(ChannelBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        try:
            type_name = CHANNEL_THREAD_NAMES[self.type]
        except KeyError:
            type_name = repr(self.type)
        
        if (type_name is not None):
            repr_parts.append(' (')
            repr_parts.append(type_name)
            repr_parts.append(')')
            
        repr_parts.append(' id=')
        repr_parts.append(repr(self.id))
        repr_parts.append(', name=')
        repr_parts.append(repr(self._get_processed_name()))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @property
    def owner(self):
        """
        Returns the thread threads's creator.
        
        Returns
        -------
        owner : ``UserClientBase``
            If user the guild has no owner, returns `ZEROUSER`.
        """
        owner_id = self.owner_id
        if owner_id:
            owner = create_partial_user_from_id(owner_id)
        else:
            owner = ZEROUSER
        
        return owner
    
    
    def is_announcements(self):
        """
        Returns whether the thread channel is bound to an announcements channel.
        
        Returns
        -------
        is_announcements : `bool`
        """
        return (self.type == 10)
    
    
    def is_public(self):
        """
        Returns whether the thread channel is public.
        
        Returns
        -------
        is_public : `bool`
        """
        return (self.type == 11)
    
    
    def is_private(self):
        """
        Returns whether the thread channel is private.
        
        Returns
        -------
        is_private : `bool`
        """
        return (self.type == 12)
    
    
    def _init_parent(self, data, guild_id):
        """
        Initializes the `.parent` attribute of the channel. If a channel is under the ``Guild``, and not in a parent
        (parent channels are all like these), then their `.parent` is the ``Guild`` itself.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        guild : ``Guild``
            The guild of the channel.
        """
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                guild.threads[self.id] = self
        
        self.guild_id = guild_id
        
        parent_id = data.get('parent_id', None)
        if (parent_id is None):
            parent_id = 0
        else:
            parent_id = int(parent_id)
        
        self.parent_id = parent_id
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelThread, cls)._create_empty(channel_id, channel_type, guild_id)
        self._messageable_init()
        
        self.archived = False
        self.archived_at = None
        self.auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.open = False
        self.owner_id = 0
        self.slowmode = 0
        self.type = channel_type
        self.thread_users = None
        self.invitable = True
        
        return self
    
    @property
    @copy_docs(ChannelBase.display_name)
    def display_name(self):
        return self.name.lower()
    
    
    @property
    @copy_docs(ChannelBase.users)
    def users(self):
        thread_users = self.thread_users
        if thread_users is None:
            users = []
        else:
            users = list(thread_users.values())
        
        return users
    
    
    @copy_docs(ChannelBase.iter_users)
    def iter_users(self):
        thread_users = self.thread_users
        if (thread_users is not None):
            yield from thread_users.values()
    
    
    @copy_docs(ChannelBase._update_attributes)
    def _update_attributes(self, data):
        self.name = data['name']
        slowmode = data.get('rate_limit_per_user', None)
        if slowmode is None:
            slowmode = 0
        self.slowmode = slowmode
        
        # Move to sub data
        data = data['thread_metadata']
        
        self.archived = data.get('archived', False)
        
        
        self.auto_archive_after = data['auto_archive_duration']*60
        
        archived_at_data = data.get('archive_timestamp', None)
        if archived_at_data is None:
            archived_at = None
        else:
            archived_at = timestamp_to_datetime(archived_at_data)
        self.archived_at = archived_at
        
        self.open = not data.get('locked', True)
        self.invitable = data.get('invitable', True)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-----------------------+-----------------------------------+
        | Keys                  | Values                            |
        +=======================+===================================+
        | archived              | `bool`                            |
        +-----------------------+-----------------------------------+
        | archived_at           | `None` or `datetime`              |
        +-----------------------+-----------------------------------+
        | auto_archive_after    | `int`                             |
        +-----------------------+-----------------------------------+
        | invitable             | `bool`                            |
        +-----------------------+-----------------------------------+
        | name                  | `str`                             |
        +-----------------------+-----------------------------------+
        | open                  | `bool`                            |
        +-----------------------+-----------------------------------+
        | slowmode              | `int`                             |
        +-----------------------+-----------------------------------+
        """
        old_attributes = {}
        
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        slowmode = data.get('rate_limit_per_user', None)
        if slowmode is None:
            slowmode = 0
        if self.slowmode != slowmode:
            old_attributes['slowmode'] = self.slowmode
            self.slowmode = slowmode
        
        
        # Move to sub data
        data = data['thread_metadata']
        
        
        archived = data.get('archived', False)
        if (self.archived != archived):
            old_attributes['archived'] = self.archived
            self.archived = archived
        
        
        auto_archive_after = data['auto_archive_duration']*60
        if (self.auto_archive_after != auto_archive_after):
            old_attributes['auto_archive_after'] = self.auto_archive_after
            self.auto_archive_after = auto_archive_after
        
        
        archived_at_data = data.get('archive_timestamp', None)
        if archived_at_data is None:
            archived_at = None
        else:
            archived_at = timestamp_to_datetime(archived_at_data)
        if (self.archived_at != archived_at):
            old_attributes['archived_at'] = self.archived_at
            self.archived_at = archived_at
        
        open_ = not data.get('locked', True)
        if (self.open != open_):
            old_attributes['open'] = self.open
            self.open = open_
        
        
        invitable = data.get('invitable', True)
        if (self.invitable != invitable):
            old_attributes['invitable'] = self.invitable
            self.invitable = invitable
        
        
        return old_attributes
    
    
    @copy_docs(ChannelBase._delete)
    def _delete(self):
        guild_id = self.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            try:
                del guild.threads[self.id]
            except KeyError:
                pass
        
        thread_users = self.thread_users
        if (thread_users is not None):
            self.thread_users = None
            channel_id = self.id
            for user in thread_users.values():
                thread_profiles = user.thread_profiles
                if (thread_profiles is not None):
                    try:
                        del thread_profiles[channel_id]
                    except KeyError:
                        pass
                    else:
                        if (not thread_profiles):
                            user.thread_profiles = None
    
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self, user):
        parent = self.parent
        if parent is None:
            return PERMISSION_NONE
        
        result = parent._permissions_for(user)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # text channels don't have voice permissions
        result &= PERMISSION_VOICE_DENY
        
        if self.type and (not result&PERMISSION_MASK_MANAGE_MESSAGES):
            result = result&PERMISSION_TEXT_DENY
            return Permission(result)
        
        if result&PERMISSION_MASK_SEND_MESSAGES:
            result &= PERMISSION_DENY_SEND_MESSAGES_ONLY
        else:
            result &= PERMISSION_TEXT_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelBase.permissions_for_roles)
    def permissions_for_roles(self, *roles):
        parent = self.parent
        if parent is None:
            return PERMISSION_NONE
        
        result = parent._permissions_for_roles(roles)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # text channels don't have voice permissions
        result &= PERMISSION_VOICE_DENY
        
        if self.type and (not result&PERMISSION_MASK_MANAGE_MESSAGES):
            result = result&PERMISSION_TEXT_DENY
            return Permission(result)
        
        if result&PERMISSION_MASK_SEND_MESSAGES:
            result &= PERMISSION_DENY_SEND_MESSAGES_ONLY
        else:
            result &= PERMISSION_TEXT_DENY
        
        return Permission(result)
    
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precrated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int` or `str`
            The channel's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        auto_archive_after: `int`, Optional (Keyword only)
            The channel's ``.auto_archive_after``.
        invitable : `bool`, Optional (Keyword only)
            The channel's `..invitable``.
        open : `bool`, Optional (Keyword only)
            The channel's ``.open``.
        name : `str`, Optional (Keyword only)
            The channel's ``.name``.
        slowmode : `int`, Optional (Keyword only)
            The channel's ``.slowmode``.
        type : `int`, Optional (Keyword only)
            The channel's ``.type``.
        
        Returns
        -------
        channel : ``ChannelThread``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        
        if kwargs:
            processable = []
            
            try:
                value = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(value, 'name', 1, 100)
                processable.append(('name', name))
            
            
            try:
                slowmode = kwargs.pop('slowmode')
            except KeyError:
                pass
            else:
                slowmode = preconvert_int(slowmode, 'slowmode', 0, 21600)
                processable.append(('slowmode', slowmode))
            
            
            try:
                auto_archive_after = kwargs.pop('auto_archive_after')
            except KeyError:
                pass
            else:
                auto_archive_after = preconvert_int_options(auto_archive_after, 'auto_archive_after',
                    AUTO_ARCHIVE_OPTIONS)
                processable.append(('auto_archive_after', auto_archive_after))
            
            
            try:
                type_ = kwargs.pop('type')
            except KeyError:
                pass
            else:
                type_ = preconvert_int(type_, 'type', 0, 256)
                if (type_ not in cls.REPRESENTED_TYPES):
                    raise ValueError(f'`type` should be one of: {cls.REPRESENTED_TYPES!r}')
                
                processable.append(('type', type_))
            
            
            for attribute_name in ('open', 'invitable'):
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    attribute_value = preconvert_bool(attribute_value, attribute_name)
                    processable.append((attribute_name, attribute_value))
            
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, cls.DEFAULT_TYPE, 0)
            CHANNELS[channel_id] = self
        
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self



__all__ = ('abort', 'InteractionResponse',)

import warnings

from ...backend.futures import is_coroutine_generator

from ...discord.embed import EmbedBase
from ...discord.client import Client
from ...discord.interaction import InteractionType

INTERACTION_TYPE_APPLICATION_COMMAND = InteractionType.application_command
INTERACTION_TYPE_MESSAGE_COMPONENT = InteractionType.message_component


def is_only_embed(maybe_embeds):
    """
    Checks whether the given value is a `tuple` or `list` containing only `embed-like`-s.
    
    Parameters
    ----------
    maybe_embeds : (`tuple` or `list`) of `EmbedBase` or `Any`
        The value to check whether is a `tuple` or `list` containing only `embed-like`-s.
    
    Returns
    -------
    is_only_embed : `bool`
    """
    if not isinstance(maybe_embeds, (list, tuple)):
        return False
    
    for maybe_embed in maybe_embeds:
        if not isinstance(maybe_embed, EmbedBase):
            return False
    
    return True


async def get_request_coroutines(client, interaction_event, show_for_invoking_user_only, response, is_return):
    """
    Gets request coroutine after an output from a command coroutine. Might return `None` if there is nothing to send.
    
    This function is a coroutine generator, which should be ued inside of an async for loop.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    show_for_invoking_user_only : `bool`
        Whether the response message should only be shown for the invoking user.
    response : `Any`
        Any object yielded or returned by the command coroutine.
    is_return : `bool`
        Whether the response is used in a return and we do not require response message.
    
    Yields
    -------
    request_coro : `None` or `coroutine`
    """
    interaction_event_type = interaction_event.type
    
    if (response is None):
        if interaction_event.is_unanswered():
            
            if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
                yield client.interaction_application_command_acknowledge(
                    interaction_event,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
            
            elif interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
                yield client.interaction_component_acknowledge(interaction_event)
        
        return
    
    if isinstance(response, (str, EmbedBase)) or is_only_embed(response):
        if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
            if interaction_event.is_unanswered():
                yield client.interaction_response_message_create(
                    interaction_event,
                    response,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
            
            elif interaction_event.is_deferred():
                yield client.interaction_followup_message_create(interaction_event, response)
            elif interaction_event.is_responded():
                yield client.interaction_followup_message_create(
                    interaction_event,
                    response,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
        
        elif interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
            yield client.interaction_component_message_edit(interaction_event, response)
        
        # No more cases
        return
    
    if is_coroutine_generator(response):
        response = await process_command_coroutine_generator(
            client,
            interaction_event,
            show_for_invoking_user_only,
            response,
        )
        
        async for request_coro in get_request_coroutines(
            client,
            interaction_event,
            show_for_invoking_user_only,
            response,
            is_return,
        ):
            yield request_coro
        
        return
    
    if isinstance(response, InteractionResponse):
        for request_coro in response.get_request_coroutines(
            client,
            interaction_event,
            show_for_invoking_user_only,
            is_return,
        ):
            yield request_coro
        
        return
    
    response = str(response)
    if len(response) > 2000:
        response = response[:2000]
    
    if response:
        
        if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
            if interaction_event.is_unanswered():
                yield client.interaction_response_message_create(
                    interaction_event,
                    response,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
            elif interaction_event.is_deferred():
                yield client.interaction_response_message_edit(interaction_event, response)
            elif interaction_event.is_responded():
                yield client.interaction_followup_message_create(
                    interaction_event,
                    response,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
        elif interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
            yield client.interaction_component_message_edit(interaction_event, response)
        
        # No more cases
        return
    else:
        if interaction_event.is_unanswered():
            
            if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
                yield client.interaction_response_message_create(
                    interaction_event,
                    show_for_invoking_user_only = show_for_invoking_user_only
                )
            elif interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
                yield client.interaction_component_acknowledge(interaction_event)
            
            return
    
    # No more cases
    return


async def process_command_coroutine_generator(
    client,
    interaction_event,
    show_for_invoking_user_only,
    coroutine_generator,
):
    """
    Processes a slash command coroutine generator.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    show_for_invoking_user_only : `bool`
        Whether the response message should only be shown for the invoking user.
    coroutine_generator : `CoroutineGenerator`
        A coroutine generator with will send command response.
    
    Returns
    -------
    response : `Any`
        Returned object by the coroutine generator.
    
    Raises
    ------
    BaseException
        Any exception raised by `coroutine_generator`.
    """
    response_message = None
    response_exception = None
    while True:
        if response_exception is None:
            step = coroutine_generator.asend(response_message)
        else:
            step = coroutine_generator.athrow(response_exception)
        
        try:
            response = await step
        except StopAsyncIteration as err:
            # catch `StopAsyncIteration` only if it is a new one.
            if (response_exception is not None) and (response_exception is not err):
                raise
            
            args = err.args
            if args:
                response = args[0]
            else:
                response = None
            break
        
        except InteractionAbortedError as err:
            response = err.response
            break
        
        except BaseException as err:
            if (response_exception is None) or (response_exception is not err):
                raise
            
            if isinstance(err, ConnectionError):
                return
            
            raise
        
        else:
            # We set it first, since if `get_request_coroutines` yields nothing, we would be meowed up.
            response_message = None
            response_exception = None
            
            async for request_coro in get_request_coroutines(
                client,
                interaction_event,
                show_for_invoking_user_only,
                response,
                False,
            ):
                try:
                    response_message = await request_coro
                except BaseException as err:
                    # `response_message` may have be set before with an iteration, so reset it.
                    response_message = None
                    response_exception = err
                    break
    
    
    return response


async def process_command_coroutine(client, interaction_event, show_for_invoking_user_only, coroutine):
    """
    Processes a slasher application command coroutine.
    
    If the coroutine returns or yields a string or an embed like then sends it to the respective channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    show_for_invoking_user_only : `bool`
        Whether the response message should only be shown for the invoking user.
    coroutine : `Coroutine`
        A coroutine which will produce command response.
    
    Raises
    ------
    BaseException
        Any exception raised by `coroutine`.
    """
    if is_coroutine_generator(coroutine):
        response = await process_command_coroutine_generator(
            client,
            interaction_event,
            show_for_invoking_user_only,
            coroutine,
        )
    else:
        try:
            response = await coroutine
        except InteractionAbortedError as err:
            response = err.response
    
    async for request_coro in get_request_coroutines(
        client,
        interaction_event,
        show_for_invoking_user_only,
        response,
        True,
    ):
        try:
            await request_coro
        except BaseException as err:
            if isinstance(err, ConnectionError):
                return
            
            raise


class InteractionResponse:
    """
    Rich interaction response message usable with `return` or with `yield` statements.
    
    May pass it's parameters to any oft he following method depending on control flow.
    
    - ``Client.interaction_response_message_create``
    - ``Client.interaction_response_message_edit``
    - ``Client.interaction_followup_message_create``
    - ``Client.interaction_component_acknowledge``
    - ``Client.interaction_component_message_edit``
    
    Attributes
    ----------
    _event : `None` or ``InteractionEvent``
        The interaction event to use instead of the default one.
    _is_abort : `bool`
        Whether the slash response is derived from an ``abort`` call.
    _message : `Ellipsis`, `None` or ``Message`` instance.
        Whether a message should be edited instead of creating a new one.
    _parameters : `dict` of (`str`, `Any`) items
        Parameters to pass to the respective ``Client`` functions.
        
        Can have the following keys:
        
        - `'allowed_mentions'`
        - `'content'`
        - `'components'`
        - `'embed'`
        - `'file'`
        - `'show_for_invoking_user_only'`
        - `'tts'`
    """
    __slots__ = ('_event', '_is_abort', '_message', '_parameters',)
    
    def __init__(self, content=..., *, embed=..., file=..., allowed_mentions=..., components=..., tts=...,
            show_for_invoking_user_only=..., message=..., event=None):
        """
        Creates a new ``InteractionResponse`` instance with the given parameters.
        
        Parameters
        ----------
        content : `str`, ``EmbedBase``, `Any`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str` or ``EmbedBase`` instance is given, then will be casted to string.
            
            If given as ``EmbedBase`` instance, then is sent as the message's embed.
            
        embed : ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `TypeError` is raised.
        file : `Any`, Optional (Keyword only)
            A file to send. Check ``Client._create_file_form`` for details.
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``Client._parse_allowed_mentions`` for details.
        components : `None`, ``ComponentBase``, (`set`, `list`) of ``ComponentBase``, Optional (Keyword only)
            Components attached to the message.
        tts : `bool`, Optional (Keyword only)
            Whether the message is text-to-speech.
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user. Defaults to the value passed when adding
            the command.
            
            If given as `True` only the message's content will be processed by Discord.
        
        message : `None`, ``Message``, Optional (Keyword only)
            Whether the interaction's message should be edited.
        event : `None`, ``InteractionEvent``, Optional (Keyword only)
            A specific event ot answer instead of the command's.
        """
        self._is_abort = False
        self._parameters = parameters = {}
        self._message = message
        self._event = event
        
        if (content is not ...):
            parameters['content'] = content
        
        if (embed is not ...):
            parameters['embed'] = embed
        
        if (file is not ...):
            parameters['file'] = file
        
        if (allowed_mentions is not ...):
            parameters['allowed_mentions'] = allowed_mentions
        
        if (components is not ...):
            parameters['components'] = components
        
        if (tts is not ...):
            parameters['tts'] = tts
        
        if (show_for_invoking_user_only is not ...):
            parameters['show_for_invoking_user_only'] = show_for_invoking_user_only
    
    def _get_response_parameters(self, allowed_parameters):
        """
        Gets response parameters to pass to a ``Client`` method.
        
        Parameters
        ----------
        allowed_parameters : `tuple` of `str`
            Allowed parameters to be passed to the respective client method.
        
        Returns
        -------
        response_parameters : `dict` of (`str`, `Any`) items
            Parameters to pass the the respective client method.
        """
        parameters = self._parameters
        response_parameters = {}
        for key in allowed_parameters:
            try:
                value = parameters[key]
            except KeyError:
                continue
            
            response_parameters[key] = value
        
        return response_parameters
    
    def get_request_coroutines(self, client, interaction_event, show_for_invoking_user_only, is_return):
        """
        Gets request coroutine buildable from the ``InteractionResponse``.
        
        This method is a generator, which should be used inside of a `for` loop.
        
        client : ``Client``
            The client who will send the responses if applicable.
        interaction_event : ``InteractionEvent``
            The respective event to respond on.
        show_for_invoking_user_only : `bool`
            Whether the response message should only be shown for the invoking user.
        is_return : `bool`
            Whether the response is used in a return and we do not require response message.
        
        Yields
        -------
        request_coro : `None` or `coroutine`
        """
        event = self._event
        if (event is not None):
            interaction_event = event
        
        interaction_event_type = interaction_event.type
        if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
            message = self._message
            if message is not ...:
                response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'components',
                    'embed', 'file'))
                
                if message is None:
                    yield client.interaction_response_message_edit(interaction_event, **response_parameters)
                else:
                    yield client.interaction_followup_message_edit(interaction_event, message, **response_parameters)
                return
            
            show_for_invoking_user_only = self._parameters.get(
                'show_for_invoking_user_only',
                show_for_invoking_user_only,
            )
            
            if ('file' in self._parameters):
                need_acknowledging = True
            elif self._is_abort:
                need_acknowledging = False
            elif is_return:
                need_acknowledging = False
            elif interaction_event.is_unanswered():
                need_acknowledging = True
            else:
                need_acknowledging = False
            
            if need_acknowledging:
                yield client.interaction_response_message_create(
                    interaction_event,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
            
            response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed',
                'file', 'tts', 'components'))
            
            if need_acknowledging or (not interaction_event.is_unanswered()):
                yield client.interaction_followup_message_create(interaction_event, **response_parameters)
            
            else:
                yield client.interaction_response_message_create(
                    interaction_event,
                    **response_parameters,
                    show_for_invoking_user_only = show_for_invoking_user_only,
                )
            
            return
        
        elif interaction_event.type is INTERACTION_TYPE_MESSAGE_COMPONENT:
            response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed', 'components'))
            if response_parameters:
                yield client.interaction_component_message_edit(interaction_event, **response_parameters)
            else:
                yield client.interaction_component_acknowledge(interaction_event)
            
            return
        
        # no more cases
    
    def __repr__(self):
        """Returns the slash response's representation."""
        repr_parts = ['<', self.__class__.__name__, ' ']
        if self._is_abort:
            repr_parts.append('(abort) ')
        
        parameters = self._parameters
        if parameters:
            for key, value in parameters.items():
                repr_parts.append(key)
                repr_parts.append('=')
                repr_parts.append(repr(value))
                repr_parts.append(', ')
            
            repr_parts[-1] = '>'
        else:
            repr_parts.append('>')
        
        return ''.join(repr_parts)


def abort(content=..., *, embed=..., file=..., allowed_mentions=..., components=..., tts=...,
        show_for_invoking_user_only=..., message=..., event=None):
    """
    Aborts the slash response with sending the passed parameters as a response.
    
    The abortion auto detects `show_for_invoking_user_only` if not given. Not follows the command's preference.
    If only a string `content` is given, `show_for_invoking_user_only` will become `True`, else `False`. The reason of
    becoming `False` at that case is, Discord ignores every other field except string content.
    
    Parameters
    ----------
    content : `str`, ``EmbedBase``, `Any`, Optional
        The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
        if any other non `str` or ``EmbedBase`` instance is given, then will be casted to string.
        
        If given as ``EmbedBase`` instance, then is sent as the message's embed.
    embed : ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
        The embedded content of the message.
        
        If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `TypeError` is raised.
    file : `Any`, Optional (Keyword only)
        A file to send. Check ``Client._create_file_form`` for details.
    allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
            , Optional (Keyword only)
        Which user or role can the message ping (or everyone). Check ``Client._parse_allowed_mentions`` for details.
    components : `None`, ``ComponentBase``, (`set`, `list`) of ``ComponentBase``, Optional (Keyword only)
        Components attached to the message.
    tts : `bool`, Optional (Keyword only)
        Whether the message is text-to-speech.
    show_for_invoking_user_only : `bool`, Optional (Keyword only)
        Whether the sent message should only be shown to the invoking user.
        
        If given as `True`, only the message's content and embeds and components will be processed by Discord.
    
    Raises
    ------
    InteractionAbortedError
        The exception which aborts the interaction, then yields the response.
    message : `None`, ``Message``, Optional (Keyword only)
        Whether the interaction's message should be edited.
    event : `None`, ``InteractionEvent``, Optional (Keyword only)
        A specific event ot answer instead of the command's.
    """
    if show_for_invoking_user_only is ...:
        if (file is not ...):
            show_for_invoking_user_only = False
        else:
            show_for_invoking_user_only = True
    
    response = InteractionResponse(content, embed=embed, file=file, allowed_mentions=allowed_mentions, components=components,
        tts=tts, show_for_invoking_user_only=show_for_invoking_user_only, message=message, event=event)
    response._is_abort = True
    raise InteractionAbortedError(response)


class InteractionAbortedError(BaseException):
    """
    An ``InteractionAbortedError`` is raised when a slash command is aborted. This class holds the response to send to
    the client.
    
    Attributes
    ----------
    response : ``InteractionResponse``
        The response to send.
    """
    def __init__(self, response):
        """
        Creates a new ``InteractionAbortedError`` instance with the given response.
        
        Parameters
        ----------
        response : ``InteractionResponse``
            The response to send.
        """
        self.response = response
        BaseException.__init__(self, response)
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'{self.__class__.__name__}({self.response!r})'


async def process_auto_completer_coroutine(client, interaction_event, coroutine):
    """
    Processes a slasher application command coroutine.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    coroutine : `Coroutine`
        A coroutine which will produce command response.
    
    Raises
    ------
    BaseException
        Any exception raised by `coroutine`.
    """
    response = await coroutine
    await client.interaction_application_command_autocomplete(interaction_event, response)
